@@ -17,7 +17,6 @@
 #include <QWaylandClientExtension>
 #include <QWindow>
 #include <QtWaylandClientVersion>
-#include <qpa/qplatformnativeinterface.h>
 
 #include <errno.h>
 #include <poll.h>
@@ -147,7 +146,9 @@ public:
 protected:
     void zwlr_data_control_offer_v1_offer(const QString &mime_type) override
     {
-        m_receivedFormats << mime_type;
+        if (!m_receivedFormats.contains(mime_type)) {
+            m_receivedFormats << mime_type;
+        }
     }
 
     QVariant retrieveData(const QString &mimeType, QMetaType type) const override;
@@ -206,8 +207,9 @@ QVariant DataControlOffer::retrieveData(const QString &mimeType, QMetaType type)
      * However this isn't actually any worse than X.
      */
 
-    QPlatformNativeInterface *native = qGuiApp->platformNativeInterface();
-    auto display = static_cast<struct ::wl_display *>(native->nativeResourceForIntegration("wl_display"));
+    auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
+    auto display = waylandApp->display();
+
     wl_display_flush(display);
 
     QFile readPipe;
@@ -221,6 +223,20 @@ QVariant DataControlOffer::retrieveData(const QString &mimeType, QMetaType type)
                 if (!img.isNull()) {
                     return img;
                 }
+            } else if (data.size() > 1 && mimeType == u"text/uri-list") {
+                const auto urls = data.split('\n');
+                QVariantList list;
+                list.reserve(urls.size());
+                for (const QByteArray &s : urls) {
+#if QT_VERSION >= QT_VERSION_CHECK(6, 7, 0)
+                    if (QUrl url(QUrl::fromEncoded(QByteArrayView(s).trimmed())); url.isValid()) {
+#else
+                    if (QUrl url(QUrl::fromEncoded(QByteArrayView(s).trimmed().toByteArray())); url.isValid()) {
+#endif
+                        list.emplace_back(std::move(url));
+                    }
+                }
+                return list;
             }
             return data;
         }
@@ -475,8 +491,8 @@ public:
         : QWaylandClientExtensionTemplate(5)
     {
         initialize();
-        auto native = qGuiApp->platformNativeInterface();
-        auto display = static_cast<struct ::wl_display *>(native->nativeResourceForIntegration("wl_display"));
+        auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
+        auto display = waylandApp->display();
         // so we get capabilities
         wl_display_roundtrip(display);
     }
@@ -541,11 +557,12 @@ WaylandClipboard::WaylandClipboard(QObject *parent)
 {
     connect(m_manager.get(), &DataControlDeviceManager::activeChanged, this, [this]() {
         if (m_manager->isActive()) {
-            QPlatformNativeInterface *native = qApp->platformNativeInterface();
-            if (!native) {
+            auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
+            if (!waylandApp) {
                 return;
             }
-            auto seat = static_cast<struct ::wl_seat *>(native->nativeResourceForIntegration("wl_seat"));
+            auto seat = waylandApp->seat();
+
             if (!seat) {
                 return;
             }
@@ -593,13 +610,17 @@ void WaylandClipboard::setMimeData(QMimeData *mime, QClipboard::Mode mode)
     }
 
     // roundtrip to have accurate focus state when losing focus but setting mime data before processing wayland events.
-    auto native = qGuiApp->platformNativeInterface();
-    auto display = static_cast<struct ::wl_display *>(native->nativeResourceForIntegration("wl_display"));
+    auto waylandApp = qGuiApp->nativeInterface<QNativeInterface::QWaylandApplication>();
+    auto display = waylandApp->display();
     wl_display_roundtrip(display);
 
     // If the application is focused, use the normal mechanism so a future paste will not deadlock itselfs
     if (m_keyboardFocusWatcher->hasFocus()) {
         QGuiApplication::clipboard()->setMimeData(mime, mode);
+        // if we short-circuit the wlr_data_device, when we receive the data
+        // we cannot identify ourselves as the owner
+        // because of that we act like it's a synchronous action to not confuse klipper.
+        wl_display_roundtrip(display);
         return;
     }
     // If not, set the clipboard once the app receives focus to avoid the deadlock
@@ -618,12 +639,12 @@ void WaylandClipboard::gainedFocus()
     // QClipboard takes ownership of the QMimeData so we need to transfer and unset our selections
     if (auto &selection = m_device->m_selection) {
         std::unique_ptr<QMimeData> data = selection->releaseMimeData();
-        WaylandClipboard::clear(QClipboard::Clipboard);
+        selection.reset();
         QGuiApplication::clipboard()->setMimeData(data.release(), QClipboard::Clipboard);
     }
     if (auto &primarySelection = m_device->m_primarySelection) {
         std::unique_ptr<QMimeData> data = primarySelection->releaseMimeData();
-        WaylandClipboard::clear(QClipboard::Selection);
+        primarySelection.reset();
         QGuiApplication::clipboard()->setMimeData(data.release(), QClipboard::Selection);
     }
 }
